Turninator = LibStub("AceAddon-3.0"):NewAddon("Turninator", "AceConsole-3.0", "AceEvent-3.0", "AceHook-3.0", "AceTimer-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("Turninator")

--==== set upvalues ====--
-- TODO: find out if QuestInfoFrame can be upvalued
-- GLOBALS: LibStub, CreateFrame, QuestInfoFrame, UIParent
local _G, strmatch, tsort = _G, string.match, table.sort
local Turninator, ceil, strfind, format, getmetatable, ipairs, next, pairs, print, select, setmetatable, strlen, strtrim, tinsert, tonumber, type, unpack = 
      Turninator, ceil, strfind, format, getmetatable, ipairs, next, pairs, print, select, setmetatable, strlen, strtrim, tinsert, tonumber, type, unpack
local AcceptQuest, AutoQuestPopUpTracker_OnMouseDown, CloseGossip, CloseQuest, CompleteQuest, DeclineQuest, GetActiveTitle, GetAutoQuestPopUp, GetGossipActiveQuests, GetGossipAvailableQuests, GetGossipOptions, GetNumActiveQuests, GetNumAutoQuestPopUps, GetNumAvailableQuests, GetNumGossipAvailableQuests, GetNumGossipOptions, GetNumQuestChoices, GetQuestReward, IsQuestCompletable, QuestChooseRewardError, SelectActiveQuest, SelectAvailableQuest, SelectGossipActiveQuest, SelectGossipAvailableQuest, SelectGossipOption = 
      AcceptQuest, AutoQuestPopUpTracker_OnMouseDown, CloseGossip, CloseQuest, CompleteQuest, DeclineQuest, GetActiveTitle, GetAutoQuestPopUp, GetGossipActiveQuests, GetGossipAvailableQuests, GetGossipOptions, GetNumActiveQuests, GetNumAutoQuestPopUps, GetNumAvailableQuests, GetNumGossipAvailableQuests, GetNumGossipOptions, GetNumQuestChoices, GetQuestReward, IsQuestCompletable, QuestChooseRewardError, SelectActiveQuest, SelectAvailableQuest, SelectGossipActiveQuest, SelectGossipAvailableQuest, SelectGossipOption
local GetBindingByKey, GetBindingKey, GetItemCount, GetItemInfo, IsInInstance, IsUsableItem, SetBinding, UnitAffectingCombat, UnitCanAttack, UnitExists, UnitGUID, UnitIsCorpse, UnitIsDead, UnitIsFriend, UnitName, UnitPlayerControlled = 
      GetBindingByKey, GetBindingKey, GetItemCount, GetItemInfo, IsInInstance, IsUsableItem, SetBinding, UnitAffectingCombat, UnitCanAttack, UnitExists, UnitGUID, UnitIsCorpse, UnitIsDead, UnitIsFriend, UnitName, UnitPlayerControlled


Turninator.DL = {
	[0] = "standard",
	"error",
	"test",
	"config",
	"events",
	"spam",
}
local DL = Turninator.DL
do	-- create fast lookups in table for reverse lookup and index number as string
	for i=0, #DL do
		DL[ DL[i] ] = i
		DL[tostring(i)] = i
	end
end

local dbgprint = { print   = function(...) return Turninator:Print(...) end,
                   squelch = function(...) return nil end }
local function SetDebugLevel(iDL)
	for i = 0,     iDL do dbgprint[ DL[i] ] = dbgprint.print;   end
	for i = iDL+1, #DL do dbgprint[ DL[i] ] = dbgprint.squelch; end
end

local tblCommands = { 
	DialogAction  = { strCommand = "CLICK TurninatorDialogActionButton:LeftButton", strConfigMsg = "taking binding for dialog event" },
	PopUpConfirm  = { strCommand = "CLICK StaticPopup1Button1:LeftButton", strConfigMsg = "taking binding for confirmation" },
	PopUpQuest    = { strCommand = "CLICK TurninatorPopUpQuestButton:LeftButton", strConfigMsg = "taking binding for quest popup" },
	OrderShipment = { strCommand = "CLICK TurninatorOrderShipmentButton:LeftButton", strConfigMsg = "taking binding to place work order" },
	-- OrderShipment = { strCommand = "CLICK GarrisonCapacitiveDisplayFrame.StartWorkOrderButton:LeftButton", strConfigMsg = "taking binding to place work order" },
	CloseShipment = { strCommand = "CLICK GarrisonCapacitiveDisplayFrameCloseButton:LeftButton", strConfigMsg = "taking binding to close work order dialog" },
	-- GarrisonCapacitiveDisplayFrame.StartWorkOrderButton
--[==[  -- TODO  LOOT
	LootConfirm   = { strCommand = "CLICK StaticPopup1Button1:LeftButton", strConfigMsg = "taking binding for loot confirm" },
--]==]  -- TODO  LOOT
	UseItem       = { strCommand = nil, strConfigMsg = "taking binding for item use" },
	Interact      = { strCommand = "INTERACTTARGET", strConfigMsg = "taking binding for interaction" },
	Released      = { strCommand = nil, strConfigMsg = "releasing binding" },
}

local CEventHook = { module = Turninator, isAttached = false }
CEventHook.__index = CEventHook
function CEventHook:Set(fAttach)
	if self.isAttached ~= fAttach then
		if fAttach then
			self.Attach(self.module, self.name, self.handler)
		else
			self.Detach(self.module, self.name)
		end
		self.isAttached = fAttach
	end
end
local CHook = { module = Turninator, isAttached = false }
CHook.__index = CHook
function CHook:Set(fAttach)
	if self.isAttached ~= fAttach then
		if fAttach then
			self.module:SecureHook(self.object, self.name, self.handler)
		else
			self.module:Unhook(self.object, self.name)
		end
		self.isAttached = fAttach
	end
end

local function CreateEventHook(type, name, handler)
	if type == "Event" then
		return setmetatable( { name = name, handler = handler or name, Attach = CEventHook.module.RegisterEvent, Detach = CEventHook.module.UnregisterEvent }, CEventHook )
	elseif type == "Hook" then
		return setmetatable( { name = name, handler = handler or name, Attach = CEventHook.module.SecureHook, Detach = CEventHook.module.Unhook }, CEventHook )
	end
end

local tblMetaEventHookSet = {}
tblMetaEventHookSet.__index = tblMetaEventHookSet
function tblMetaEventHookSet:Set(value)
	for i=1, #self do
		self[i]:Set(value)
	end
end


local tblMainEventHooks = setmetatable( {
	CreateEventHook("Event", "GOSSIP_SHOW",           "DialogEventHandler"),
	CreateEventHook("Event", "GOSSIP_CLOSED",         "DialogEventHandler"),
	CreateEventHook("Event", "QUEST_GREETING",        "DialogEventHandler"),
	CreateEventHook("Event", "QUEST_DETAIL",          "DialogEventHandler"),
	CreateEventHook("Event", "QUEST_PROGRESS",        "DialogEventHandler"),
	CreateEventHook("Event", "QUEST_COMPLETE",        "DialogEventHandler"),
	CreateEventHook("Event", "QUEST_FINISHED",        "DialogEventHandler"),
	CreateEventHook("Event", "MERCHANT_SHOW",         "MerchantEventHandler"),
	CreateEventHook("Event", "MERCHANT_CLOSED",       "MerchantEventHandler"),
	CreateEventHook("Event", "PLAYER_TARGET_CHANGED", "SelectBinding"),
--	CreateEventHook("Event", "UPDATE_BINDINGS",       "UpdateBindingsHandler"),
--	CreateEventHook("Hook",  "AddAutoQuestPopUp",     "AutoQuestPopupEventHandler"),
--	CreateEventHook("Hook",  "RemoveAutoQuestPopUp",  "AutoQuestPopupEventHandler"),
	CreateEventHook("Hook",  "AutoQuestPopupTracker_AddPopUp",     "SelectBinding"),
	CreateEventHook("Hook",  "AutoQuestPopupTracker_RemovePopUp",  "SelectBinding"),
	
	CreateEventHook("Event", "SHIPMENT_CRAFTER_INFO",    "SelectBinding"),
	CreateEventHook("Event", "SHIPMENT_CRAFTER_CLOSED",  "SelectBinding"),
	-- CreateEventHook("Hook",  "RequestShipmentCreation", "Print"),
	-- setmetatable( { object = C_Garrison, name = "RequestShipmentCreation", handler = "SelectBinding", }, CHook )

--[==[  -- TODO  LOOT
	CreateEventHook("Event", "LOOT_BIND_CONFIRM",     "LootConfirm"),
--	CreateEventHook("Event", "LOOT_OPENED",           "LootOpened"),
	CreateEventHook("Event", "LOOT_READY",            "LootOpened"),
	CreateEventHook("Event", "LOOT_CLOSED",           "LootClosed"),
	CreateEventHook("Event", "LOOT_SLOT_CLEARED",     "LootSlotCleared"),
	CreateEventHook("Event", "LOOT_SLOT_CHANGED",     "LootSlotChanged"),
--]==]  -- TODO  LOOT
}, tblMetaEventHookSet )


local tblZoneEventHooks = setmetatable( {
	CreateEventHook("Event", "ZONE_CHANGED_NEW_AREA", "UpdateHooks"),
	CreateEventHook("Event", "PLAYER_ENTERING_WORLD", "UpdateHooks"),
}, tblMetaEventHookSet )

local tblPopupEventHooks = setmetatable( {
	CreateEventHook("Hook",  "StaticPopup_Show",  "StaticPopupShow"),
	CreateEventHook("Hook",  "StaticPopup_Hide",  "StaticPopupHide"),
}, tblMetaEventHookSet )

local salvageEventHook = CreateEventHook("Event", "BAG_UPDATE", "SalvageHandler")
-- local tradingEventHook = CreateEventHook("Event", "BAG_UPDATE", "TradingHandler")


--[==[  -- TODO  LOOT
local function HandleLootClick(tblLoot)
	local strFrame = StaticPopup_Visible("LOOT_BIND")
	if strFrame ~= nil then
		_G[strFrame.."Button1"]:Click("LeftButton", false)
	else
		local slot = tblLoot and next(tblLoot)
		if slot then LootSlot(slot) end
	end
end

local tblLoot = {}

function Turninator:LootConfirm(event, slot, ...)
	if GetNumGroupMembers() == 0 then
		if type(slot) ~= "number" or GetNumLootItems() < slot or not LootSlotHasItem(slot) then  -- sanity check, probably not needed
			dbgprint.error("sanity check failed", event, slot, ...)
		else
			dbgprint.test(event, slot, ...)
			tblLoot[slot] = true
			self.tblDialogAction = { HandleLootClick, tblLoot } 
			self:SelectBinding(event);
		end	
	end
end

function Turninator:LootSlotCleared(event, slot, ...)
	dbgprint.test(event, slot, ...)
	tblLoot[slot] = nil
end

function Turninator:LootSlotChanged(event, ...)
	dbgprint.test(event, ...)
end

function Turninator:LootOpened(event, ...)
	dbgprint.test(event, ...)
end

function Turninator:LootClosed(event, ...)
	dbgprint.test(event, ...)
	tblLoot = {}
	self.tblDialogAction = {} 
	self:SelectBinding(event)
end
--]==]  -- TODO  LOOT


-- Called when the addon is loaded
function Turninator:OnInitialize()
	Turninator:LoadSavedSettings()

	LibStub("AceConfig-3.0"):RegisterOptionsTable("Turninator", Turninator.GetUIOptions, { "Turninator", "tr" } )
	self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("Turninator", "Turninator")

	for k,v in pairs(self.DL) do DL[v] = k; DL[k] = v end
	SetDebugLevel(self.db.global.iDebugLevel)

	-- if there's no shared command, then have the binding release to Interact with Target
	tblCommands.Released.strCommand = self.db.profile.strSharedCmd or tblCommands.Interact.strCommand
	self.strCurBind = "Released"
	self.tblDialogAction = {}
end

function Turninator:UpdateHooks(fForce)
	local fHookMains = self.db.profile.fEnabled and not (self.db.profile.fDisableInInstances and IsInInstance())
	dbgprint.spam(fHookMains and "enabling addon" or "disabling addon")
	tblMainEventHooks:Set(fHookMains)

	local fHookZones = self.db.profile.fEnabled and self.db.profile.fDisableInInstances
	tblZoneEventHooks:Set(fHookZones)

	self:SelectBinding("UpdateHooks")
end

-- Called when the addon is enabled
function Turninator:OnEnable()
	self:CreateButtons()
	self:UpdateHooks()
end

-- Called when the addon is Disabled
function Turninator:OnDisable()
	if (self.db.profile.strSharedKey) then
		SetBinding(self.db.profile.strSharedKey, self.db.profile.strSharedCmd);
	end

	tblMainEventHooks:Set(false)
	tblZoneEventHooks:Set(false)
end

function Turninator:SetEnabledState(info, value)
	value = value or (info and info[#info])
	if (value == L["On"]) then
		self.db.profile.fEnabled = true
		self:UpdateHooks()
	elseif (value == L["Off"]) then
		self.db.profile.fEnabled = false
		self.tblDialogAction = {}
		self:UpdateHooks()
	else
		dbgprint.standard("ERROR: no valid state passed to SetEnabledState");
	end
	
	self:Config(info)
end

function Turninator:CreateButtons()
	if (not self.btns) then
		local DialogAction, PopUpQuest, OrderShipment

		DialogAction = CreateFrame("Button", "TurninatorDialogActionButton", UIParent, "SecureActionButtonTemplate");
		DialogAction:SetAttribute("type", "macro")
		DialogAction:SetAttribute("macrotext", "/run Turninator:DialogAction()")

		PopUpQuest = CreateFrame("Button", "TurninatorPopUpQuestButton", UIParent, "SecureActionButtonTemplate");
		PopUpQuest:SetAttribute("type", "macro")
		PopUpQuest:SetAttribute("macrotext", "/run Turninator:PopUpQuest()")

		OrderShipment = CreateFrame("Button", "TurninatorOrderShipmentButton", UIParent, "SecureActionButtonTemplate");
		OrderShipment:SetAttribute("type", "macro")
		OrderShipment:SetAttribute("macrotext", "/run Turninator:OrderShipment()")

		self.btns = { DialogAction, PopUpQuest, OrderShipment }
	end
end


function Turninator:EventHandler(...)
	dbgprint.events(...)
end

function Turninator:UpdateBindingsHandler(...)
	local strBind = self.db.profile.strSharedKey and GetBindingByKey(self.db.profile.strSharedKey) or ""
	dbgprint.events(..., strBind)
end

function Turninator:DialogEventHandler(event, ...)
	dbgprint.events(event, ...)
	self.tblDialogAction = self:GetDialogAction(event, ...)
	self:SelectBinding(event);
end

Turninator.tblConfirmMerchant = { [86683] = true, [86776] = true, [86777] = true, [86778] = true, [86779] = true, [87200] = true, [87201] = true, [87202] = true, [87203] = true, [87204] = true, [96130] = true }
-- static popup events are problematic. will need to be fixed should we ever have vendor like these where we want to confirm and useitem with same npc
-- best solution there might just be to tell turn those npcs into useitem npcs; would lose using button to interact but that might be okay
-- Turninator.tblTradingPostMerchant = { [86683] = true, [86776] = true, [86777] = true, [86778] = true, [86779] = true, [87200] = true, [87201] = true, [87202] = true, [87203] = true, [87204] = true }
Turninator.tblSalvageMerchant = { [77378] = true, [79857] = true }
Turninator.tblSalvageItems = { 139593, 140590, 114116, 114119, 114120 }

function Turninator:SalvageHandler(event, ...)
	-- dbgprint.test(event, ...)
	for k,id in ipairs(self.tblSalvageItems) do
		-- dbgprint.test(k, id, GetItemCount(id), (GetItemInfo(id)))
		if GetItemCount(id) > 0 then
			if id ~= self.idCurrentSalvage then
				self.idCurrentSalvage = id
				self.db.global.tblUseItem[self.strSalvageMerchant] = GetItemInfo(id)
				self:TakeBinding("Released", "Salvage")
				self:SelectBinding("UseItem")
			end
			return
		end
	end
end

function Turninator:MerchantEventHandler(event, ...)
	dbgprint.events(event, ...)
	if event == "MERCHANT_CLOSED" then
		if self.fConfirmMerchant then
			self.fConfirmMerchant = nil
			self:SelectBinding(event);
--[===[ TODO - figure out what the deal is with static popups
		elseif self.strTradingPostMerchant then
			tblPopupEventHooks:Set(false)
			self.db.global.tblUseItem[self.strTradingPostMerchant] = nil
			self.strTradingPostMerchant = nil
			self:SelectBinding(event);
--]===]
		elseif self.strSalvageMerchant then
			salvageEventHook:Set(false)
			self.db.global.tblUseItem[self.strSalvageMerchant] = nil
			self.strSalvageMerchant = nil
			self.idCurrentSalvage = nil
			self:SelectBinding(event);
		end
	elseif event == "MERCHANT_SHOW" then
		local npc = tonumber(strmatch(UnitGUID("target"), "Creature%-0%-%w+%-%w+%-%w+%-(%w+)%-%w+"))
		if self.tblConfirmMerchant[npc] then
			self.fConfirmMerchant = true
			self:SelectBinding(event);
--[===[ TODO - figure out what the deal is with static popups
		elseif self.tblTradingPostMerchant[npc] then
			self.strTradingPostMerchant = UnitName("target")
			self.db.global.tblUseItem[self.strTradingPostMerchant] = GetItemInfo(120146) -- Smuggled Sack of Gold
			tblPopupEventHooks:Set(true)
			self:TakeBinding("Released", "Trading Post")
			self:SelectBinding("UseItem")
--]===]
		elseif self.tblSalvageMerchant[npc] then
			self.strSalvageMerchant = UnitName("target")
			salvageEventHook:Set(true)
			self:SalvageHandler(event)
		end
	end
end

--[===[
local LootSlot = {} -- added because otherwise we'd be referencing a global table below ...
local function HandleLootClick(tblLoot)
	local strFrame = StaticPopup_Visible("LOOT_BIND")
	if strFrame ~= nil then
		_G[strFrame.."Button1"]:Click("LeftButton", false)
	else
		local slot = tblLoot and next(tblLoot)
		if slot then LootSlot(slot) end  -- what the heck was this supposed to accomplish? was that our table or the games? confused!
	end
end
--]===]

local tblLoot = {}

function Turninator:StaticPopupShow(name, ...)
	dbgprint.events("popup show", name, ...)
	-- dbgprint.test("old bind", self.strCurBind)
	self.fConfirmMerchant = true
	--self.tblDialogAction = { HandleLootClick, tblLoot } 
	self:SelectBinding("popup show "..name);
end

-- TODO - this check isn't working right. We can get a name like CONFIRM_PURCHASE_TOKEN_ITEM
-- from PopupShow hook, but then we get multiple calls here with EQUIP_BIND as name if we 
-- mouse around over stuff. And then we DON'T get a hide for CONFIRM_PURCHASE_TOKEN_ITEM so
-- we can't just cache the name from show and wait for matching hide. Best bet may be to try
-- hooking the frame once its up to know when it goes away.
function Turninator:StaticPopupHide(name, ...)
	dbgprint.events("popup hide", name, ...)
	self.fConfirmMerchant = false
	--self.tblDialogAction = { HandleLootClick, tblLoot } 
	self:SelectBinding("popup hide "..name);
	-- dbgprint.test("new bind", self.strCurBind)
end

function Turninator:ShipmentEventHandler(event, ...)
	if #self.tblDialogAction > 1 then
		self.tblDialogAction = {}
	end

	if (event == "SHIPMENT_CRAFTER_INFO") then
		local action
		local gcdf = _G.GarrisonCapacitiveDisplayFrame
		if (gcdf and gcdf.StartWorkOrderButton and gcdf.StartWorkOrderButton:IsVisible() and gcdf.StartWorkOrderButton:IsEnabled()) then
			action = _G.C_Garrison.RequestShipmentCreation
		end
		if self.tblDialogAction[1] ~= action then
			dbgprint.events(event, ...)
			self.tblDialogAction[1] = action
		end

	elseif (event == "SHIPMENT_CRAFTER_CLOSED") then
		dbgprint.events(event, ...)
		self.tblDialogAction = {}

	end

	self:SelectBinding(event);
end

local function CanOrderShipment()
	local btn = _G.GarrisonCapacitiveDisplayFrame and _G.GarrisonCapacitiveDisplayFrame.StartWorkOrderButton
	return btn and btn:IsVisible() and btn:IsEnabled()
end

local function CanCloseShipment()
	local btn = _G.GarrisonCapacitiveDisplayFrameCloseButton
	return btn and btn:IsVisible() and btn:IsEnabled()
end

function Turninator:OrderShipment()
	_G.C_Garrison.RequestShipmentCreation()
end

function Turninator:PopUpQuest()
	for i=1, GetNumAutoQuestPopUps() do
		local questId = GetAutoQuestPopUp(i)
		local block = _G.AUTO_QUEST_POPUP_TRACKER_MODULE:GetExistingBlock(questId);
		if block then
			AutoQuestPopUpTracker_OnMouseDown(block)
			return
		end
	end
end

function Turninator:AutoQuestPopupEventHandler(event, ...)
	dbgprint.events(event, ...)
	for i=1, GetNumAutoQuestPopUps() do
		local questId = GetAutoQuestPopUp(i)
		local block = _G.AUTO_QUEST_POPUP_TRACKER_MODULE:GetExistingBlock(questId);
		if block then
			self.tblDialogAction = { AutoQuestPopUpTracker_OnMouseDown, block }
			self:SelectBinding(event);
			return
		end
	end
end

function Turninator:AddAutoQuestPopUpHook(...) -- questId, type)
	dbgprint.events("AddAutoQuestPopUpHook", ...)
	self:SelectBinding("AddAutoQuestPopUpHook")
end

function Turninator:RemoveAutoQuestPopUpHook(...) -- questId
	dbgprint.events("RemoveAutoQuestPopUpHook", ...)
	self:SelectBinding("RemoveAutoQuestPopUpHook")
end


function Turninator:Config(info)
	if (info and info.uiType == "cmd") then
		LibStub("AceConfigRegistry-3.0"):NotifyChange("Turninator")
		if (info[#info] == L["Config"]) then 
			_G.InterfaceOptionsFrame_OpenToCategory(Turninator.optionsFrame)
		else
			dbgprint.error("debug reporting level set to", self.db.global.iDebugLevel, DL[self.db.global.iDebugLevel])
			if (self.db.profile.fEnabled and self.db.profile.fDisableInInstances and IsInInstance()) then
				dbgprint.standard(L["addon is"], L["temporarily disabled for an instance"])
			else
				dbgprint.standard(L["addon is"], self.db.profile.fEnabled and L["enabled"] or L["disabled"])
			end
			dbgprint.standard(L["shared cmd is"], self.db.profile.strSharedCmd or L["not set"])
			dbgprint.standard(L["shared key is"], self.db.profile.strSharedKey or L["not set"])
			if (self.db.profile.strSharedKey) then 
				dbgprint.spam("shared key currently bound to", GetBindingByKey(self.db.profile.strSharedKey))
			end
		end
	end
end


function Turninator:Debug(info, value)
	local strValue, index = self:GetArgs(value, 1);

	if (strValue ~= nil) then
		if (index ~= 1e9) then
			dbgprint.standard(L["ERROR: too many parameters passed"]);
		end

		local iValue = DL[strValue]
		if (iValue == nil) then
			dbgprint.standard("ERROR: value must be number between 0 and", #DL);
			return;
		else
			self.db.global.iDebugLevel = iValue
			SetDebugLevel(iValue)
		end
	end

	self:Config(info)
end

function Turninator:SetIconCallback(idItem)
	local strItemName, _, _, _, _, _, _, _, _, strItemIcon = GetItemInfo(idItem)
	if not strItemName then
		dbgprint.standard("ERROR: invalid item id: ", idItem);
	else
		dbgprint.standard("second attempt successful for ", idItem);
		self.db.global.tblItemIcon[strItemName] = strItemIcon
	end
end

function Turninator:SetIconById(info, value)
	local strValue, index = self:GetArgs(value, 1, 1)
	while (strValue~= nil) do
		local idItem = tonumber(strValue)
		if not idItem then
			dbgprint.standard("ERROR: non-numeric argument passed: ", strValue);
		else
			local strItemName, _, _, _, _, _, _, _, _, strItemIcon = GetItemInfo(idItem)
			if not strItemName then
				self:ScheduleTimer("SetIconCallback", 1, idItem);
				-- dbgprint.standard("ERROR: invalid item id: ", idItem);
			else
				self.db.global.tblItemIcon[strItemName] = strItemIcon
			end
		end
		strValue, index = self:GetArgs(value, 1, index)
	end
end

function Turninator:Reset(info)
	if (self.db.profile.strSharedKey) then
		SetBinding(self.db.profile.strSharedKey, self.db.profile.strSharedCmd);
	end

	self.db.profile.strSharedKey = nil
	self.db.profile.strSharedCmd = nil

	tblCommands.Released.strCommand = tblCommands.Interact.strCommand
	self.strCurBind = "Released"
	self.tblDialogAction = {}

	self:Config(info)
end

function Turninator:ShareKey(info, value)
	local strKey, index = self:GetArgs(value, 1);

	if (strKey == nil) then
		if (info and info.uiType == "cmd") then
			dbgprint.standard(L["ERROR: no key specified"]);
		else
			Turninator:Reset(info)
		end
		return
	elseif (index ~= 1e9) then
		dbgprint.standard(L["ERROR: too many parameters passed"]);
		return
	end

	local strAction = GetBindingByKey(strKey)
	if (not SetBinding(strKey, strAction)) then
		dbgprint.standard(L["ERROR: invalid key"])
		return;
	end

	self:Reset();  -- release old binding before setting up new one

	if (not strAction or strAction == "") then
		dbgprint.config("sharing previously unbound key");
	else
		dbgprint.config("sharing key", strKey, "with action", strAction);
		self.db.profile.strSharedCmd = strAction
		tblCommands.Released.strCommand = strAction
	end
	self.db.profile.strSharedKey = strKey

	self.strCurBind = "Released"
	self:SelectBinding("ShareKey")
	self:Config(info)
end

function Turninator:ShareCmd(info, value)
	local strAction, index = self:GetArgs(value, 1);

	if (strAction == nil) then
		dbgprint.standard(L["ERROR: no action specified"]);
		return;
	elseif (index ~= 1e9) then
		dbgprint.standard(L["ERROR: too many parameters passed"]);
		return
	end

	local strKey = GetBindingKey(strAction)
	if (not strKey or strKey == "") then
		dbgprint.standard(L["ERROR: action does not exist or is unbound"]);
		return;
	end

	self:Reset();  -- release old binding before setting up new one

	self.db.profile.strSharedCmd = strAction
	self.db.profile.strSharedKey = strKey

	tblCommands.Released.strCommand = strAction
	self.strCurBind = "Released"
	self:SelectBinding("ShareCmd")
	self:Config(info)
end

function Turninator:UseItem(info, strItem)
	local strName = UnitName("target")
	if (not strName or UnitPlayerControlled("target")) then
		dbgprint.standard(L["ERROR: please target an NPC when using this command"])
		return
	elseif (strItem == nil) then
		dbgprint.standard(L["ERROR: no item specified"]);
		return
	end

	local strItemName, _, _, _, _, _, _, _, _, strItemIcon = GetItemInfo(strItem)
	if (strItemName == nil) then
		dbgprint.standard(L["ERROR: item not found in inventory"]);
		return
	end

	-- GET_ITEM_INFO_RECEIVED
	-- ITEM_PUSH

	dbgprint.standard("adding item", strItemName, "for use on", strName)
	self.db.global.tblUseItem[strName] = strItemName
	self.db.global.tblItemIcon[strItemName] = strItemIcon
	self:TakeBinding("Released", "UseItem")
	self:SelectBinding("UseItem")
end

function Turninator:SetNPC(info)
	local strAction = info and info[#info]
	local strName = UnitName("target")

	if (not strName or UnitPlayerControlled("target")) then
		dbgprint.standard(L["ERROR: please target an NPC when using this command"])
	elseif (strAction == L["Interact"]) then
		dbgprint.standard("adding", strName, "to Interact list")
		self.db.global.tblInteract[strName] = true
		self:SelectBinding("Interact")
	elseif (strAction == L["Ignore"]) then
		dbgprint.standard("adding", strName, "to Ignore list")
		self.db.global.tblInteract[strName] = false
		self:SelectBinding("Ignore")
	elseif (strAction == L["Forget"]) then
		dbgprint.standard("forgetting", strName)
		self.db.global.tblInteract[strName] = nil
		self.db.global.tblUseItem[strName] = nil
		self:SelectBinding("Forget")
	end

	LibStub("AceConfigRegistry-3.0"):NotifyChange("Turninator");
end

function Turninator:ListNPCs(info)
	local ignore, interact, items = {}, {}, {}
	local cmac = 0

	for strNPCName,fInteract in pairs(self.db.global.tblInteract) do
		tinsert( (fInteract and interact or ignore), strNPCName)
	end

	if (#interact > 0) then
		dbgprint.standard(L["forcing interaction with:"])
		tsort(interact)
		for _,v in ipairs(interact) do
			print(format("  %s", v))
		end
	end

	if (#ignore > 0) then
		dbgprint.standard(L["ignoring:"])
		tsort(ignore)
		for _,v in ipairs(ignore) do
			print(format("  %s", v))
		end
	end


	for strNPCName,strItemName in pairs(self.db.global.tblUseItem) do
		tinsert( items, { name = strNPCName, item = strItemName } )
		local clen = strlen(strNPCName)
		if (cmac < clen) then cmac = clen end
	end

	if (#items) then 
		dbgprint.standard(L["npc item bindings:"])
		tsort(items, function(a,b) return (a.name < b.name) end)
		for _,v in ipairs(items) do
			print(format( format("  %%-%is%%s", cmac+3), v.name, v.item ))
		end
	end
end


function Turninator:SelectBinding(reason)
	if (self.db.profile.strSharedKey == nil) then 
		if (self.db.profile.fEnabled) then dbgprint.error("ERROR: no key is set; cannot respond to", reason); end
	elseif (not self.db.profile.fEnabled) then
		self:TakeBinding("Released", reason)
	elseif (self.fConfirmMerchant) then
		self:TakeBinding("PopUpConfirm", reason)
	elseif (#self.tblDialogAction ~= 0) then
		self:TakeBinding("DialogAction", reason)
	elseif (GetNumAutoQuestPopUps() ~= 0) then
--		self:AutoQuestPopupreasonHandler(reason)
		self:TakeBinding("PopUpQuest", reason)
	elseif (CanOrderShipment()) then
		self:TakeBinding("OrderShipment", reason)
	elseif (CanCloseShipment()) then
		self:TakeBinding("CloseShipment", reason)
	elseif (self:CanUseItem()) then
		self:TakeBinding("UseItem", reason)
	elseif (self:CanInteract()) then
		self:TakeBinding("Interact", reason)
	else
		self:TakeBinding("Released", reason)
	end
end

function Turninator:TakeBinding(strBind, reason)
	-- print("TakeBinding called", reason, strBind)
	if (self.strCurBind ~= strBind) then
		if (UnitAffectingCombat("player")) then
			dbgprint.error("currently in combat; binding cannot be changed")
		else
			local bind = tblCommands[strBind]
			if (SetBinding(self.db.profile.strSharedKey, bind.strCommand)) then
				dbgprint.config(bind.strConfigMsg, reason);
				self.strCurBind = strBind
			end
		end
	end
end


local function MakeLocalizedTable(t, k)
	local meta = getmetatable(t)
	for itemId, count in pairs(meta.items) do
		local name = GetItemInfo(itemId)
		if name then
			t[name] = { itemId = itemId, count = count }
			meta.items[itemId] = nil
		end
	end
	if not next(meta.items, nil) then
		setmetatable(t,nil)
	end
end
Turninator.tblProblemTurnins = {
	[87706] = setmetatable( {}, { __index = MakeLocalizedTable, items = {
			[118654] = 1,  -- 		Aogexon's Fang
			[118655] = 1,  -- 		Bergruu's Horn
			[118656] = 1,  -- 	   Dekorhan's Tusk
			[118657] = 1,  -- 	   Direhoof's Hide
			[118658] = 1,  -- 		Gagrog's Skull
			[118659] = 1,  -- 		 Mu'gra's Head
			[118660] = 1,  -- 	Thek'talon's Talon
			[120172] = 1,  -- 	   Vileclaw's Claw
			[118661] = 1,  -- 	Xelganak's Stinger
		}}
	),
	[87393] = { 
				  [L["A Rare Find"]] = { itemId = 118100, count = 1  },  -- highmaul relic
		[L["Fragments of the Past"]] = { itemId = 118099, count = 20 },  -- gorian artifact fragment
	},
	[85984] = setmetatable( {}, { __index = MakeLocalizedTable, items = {
			[116818] = 1,  --      Abyssal Gulper Lunker 
			[116817] = 1,  -- Blackwater Whiptail Lunker 
			[116820] = 1,  --          Blind Lake Lunker 
			[116821] = 1,  --         Fat Sleeper Lunker 
			[127994] = 1,  --     Felmouth Frenzy Lunker 
			[116819] = 1,  --       Fire Ammonite Lunker 
			[116822] = 1,  --     Jawless Skulker Lunker 
			[122696] = 1,  --        Sea Scorpion Lunker 
		}}
	),
};

function Turninator:GetDialogAction(event, ...)

	if (event == "GOSSIP_SHOW") then
		local function GetActive(...)
			local tbl = {}
			for iOffset = 1, select("#", ...), 6 do
				local iActive = ceil(iOffset/6)
				local strQuest = select(iOffset, ...)
				local fComplete = select(iOffset+3, ...)
				tbl[iActive] = { name = strQuest, complete = fComplete }
			end
			return tbl;
		end

		local function GetAvailable(...)
			local function CompareAvailableQuests(a, b)
				if (not (a.precedence == b.precedence)) then
					return a.precedence < b.precedence
				elseif (not (a.level == b.level)) then
					return a.level > b.level
				else
					return a.index < b.index
				end
			end

			local tbl = {}
			for iOffset = 1, select("#", ...), 7 do
				local index = ceil(iOffset/7)
				local strQuest, level, isTrivial, frequency, isRepeatable, isLegendary, __ = select(iOffset, ...)
				local precedence = (isRepeatable and 100) or (isTrivial and 50) or frequency
				tinsert(tbl, { index = index, name = strQuest, level = level, precedence = precedence } )
			end
			tsort(tbl, CompareAvailableQuests)
			return tbl;
		end

		local function GetGossip(...)
			local tbl = {}
			for iOffset = 1, select("#", ...), 2 do
				local index = ceil(iOffset/2)
				local strGossip, strType = select(iOffset, ...)
				if (strType == "gossip") then
					tinsert(tbl, { index = index, text = strGossip } )
				end
			end
			return tbl;
		end

		-- check for quests we can turn in
		local tblActive = GetActive(GetGossipActiveQuests())
		for i = 1, #tblActive do
			if (tblActive[i].complete) then
				-- found a quest we can turn in, that's what button will do
				dbgprint.spam("will turn in quest", i, tblActive[i].name);
				return { SelectGossipActiveQuest, i }
			end
		end

		-- no quests we can turn in, so how about pick up an available quest
		local cAvailable = GetNumGossipAvailableQuests()
		dbgprint.spam(cAvailable, "available quests for this npc")
		if (cAvailable > 0) then
			local tblAvailable = GetAvailable(GetGossipAvailableQuests())

			local guid = UnitGUID("target")
			local npc = guid and strmatch(guid, "Creature%-0%-%w+%-%w+%-%w+%-(%w+)%-%w+")
			local turnins = npc and self.tblProblemTurnins[tonumber(npc)]
			if not turnins then
				return { SelectGossipAvailableQuest, tblAvailable[1].index }
			else
				for _, quest in ipairs(tblAvailable) do
					local turnin = turnins[quest.name]
					if turnin and turnin.count <= GetItemCount(turnin.itemId) then
						return { SelectGossipAvailableQuest, quest.index }
					end		
				end
			end
		end

		-- if the npc has chatty gossip options try the first one
		if (GetNumGossipOptions() > 0) then
			local tblGossip = GetGossip(GetGossipOptions())
			dbgprint.spam(#tblGossip, "gossip text options for this npc")
			-- PrintTable(tblGossip)
			for _, tblChoice in pairs(tblGossip) do
				if (strfind(tblChoice.text, L["Darkmoon Game Token"])) then
					return { SelectGossipOption, tblChoice.index }
				elseif (strfind(tblChoice.text, L["Take me to the faire staging area"])) then
					return { SelectGossipOption, tblChoice.index, nil, true}
				elseif (strfind(tblChoice.text, L["Teleport me to the cannon"])) then
					return { SelectGossipOption, tblChoice.index, nil, true}
				end
			end
			if (tblGossip[1] ~= nil) then 
				return { SelectGossipOption, tblGossip[1].index }
			end
		end

		-- nothing we can do with this quest giver apparently ... exit seems iffy ... revisit
		return { CloseGossip }

	elseif (event == "QUEST_GREETING") then
		local function GetActive()
			local tbl = {}
			for iActive = 1, GetNumActiveQuests() do
				local strQuest, fComplete = GetActiveTitle(iActive)
				tbl[iActive] = { name = strQuest, complete = fComplete }
			end
			return tbl;
		end

		local tblActive = GetActive()
		for i = 1, #tblActive do
			if (tblActive[i].complete) then
				-- found a quest we can turn in, that's what button will do
				dbgprint.spam("will turn in quest", i, tblActive[i].name);
				return { SelectActiveQuest, i }
			end
		end

		-- no quests we can turn in, so how about pick up an available quest
		local cAvailable = GetNumAvailableQuests()
		dbgprint.spam(cAvailable, "available quests for this npc")
		if (cAvailable > 0) then
			return { SelectAvailableQuest, 1 }
		end

		-- nothing we can do with this quest giver apparently ... exit seems iffy ... revisit
		return { CloseQuest }

	elseif (event == "QUEST_DETAIL") then
		return { AcceptQuest }

	elseif (event == "QUEST_PROGRESS") then
		if (IsQuestCompletable()) then
			return { CompleteQuest }
		else
			return { DeclineQuest }
		end
		-- todo - can't complete, back up?

	elseif (event == "QUEST_COMPLETE") then
		local function CheckRewardSelection()
			if (QuestInfoFrame.itemChoice ~= 0) then
				GetQuestReward(QuestInfoFrame.itemChoice)
			else
				QuestChooseRewardError()
			end 
		end

		local numRewardChoices = GetNumQuestChoices()
		if (numRewardChoices == 0) then
			return { GetQuestReward }
		elseif (numRewardChoices == 1) then
			return { GetQuestReward, 1 }
		else
			return { CheckRewardSelection }
		end

	elseif (event == "GOSSIP_CLOSED") then
		return {}

	elseif (event == "QUEST_FINISHED") then
		return {}
	end

	return {}
end

function Turninator:DialogAction()
	if (not self.tblDialogAction or type(self.tblDialogAction) ~= "table") then
		dbgprint.error("ERROR - tblDialogAction is of type", type(self.tblDialogAction));
	elseif (not self.tblDialogAction[1] or type(self.tblDialogAction[1]) ~= "function") then
		dbgprint.error("ERROR - tblDialogAction does not contain function");
	else
		self.tblDialogAction[1](unpack(self.tblDialogAction, 2));
	end
end

local tblItemIDFixups = {
	[74258] = 74771,  -- Staff of Pei-Zhi
}

function Turninator:CanUseItem()
	local name = UnitName("target")
	if (not name or UnitPlayerControlled("target")) then
		return false
	end
	local item = self.db.global.tblUseItem[name]

	if (not item) then
		return false
	end
	
	 if not IsUsableItem(item) then
		local _, strItemLink, _, _, _, _, _, _, _, strItemIcon = GetItemInfo(item)
		local idItem = strItemLink and strmatch(strItemLink, "item:(%d+)")
		local idFixed = tblItemIDFixups[tonumber(idItem)]
		if not idFixed or not IsUsableItem(idFixed) then
			return false
		end
	end

	-- if we're missing the icon path (association created by v1 of the addon) grab it now
	if (not self.db.global.tblItemIcon[item]) then
		local strItemName, _, _, _, _, _, _, _, _, strItemIcon = GetItemInfo(item)
		if (item == strItemName) then
			self.db.global.tblItemIcon[strItemName] = strItemIcon
		end
	end

	-- set the binding up in tblCommands
	tblCommands.UseItem.strCommand = format("ITEM %s", item)
	-- print(tblCommands.UseItem.strCommand)
	return true
end

function Turninator:CanInteract()
	local name = UnitName("target")
	if name == "Human Orphan" or name == "Draenei Orphan" then return true end

	if (not name or UnitPlayerControlled("target")) then
		return false
	end

	local interact = self.db.global.tblInteract[name]
	if (interact ~= nil) then
		return interact
	end

	return (UnitIsFriend("player", "target") and not UnitIsCorpse("target") 
		and not UnitIsDead("target") and not UnitCanAttack("player", "target"))
end

